@@ -73,6 +73,13 @@ class Residual(nn.Module):
     def forward(self, x):
         return self.fn(x) + x
 
+class Flatten(nn.Module):
+    def __init__(self, index):
+        super().__init__()
+        self.index = index
+    def forward(self, x):
+        return x.flatten(self.index)
+
 class Rezero(nn.Module):
     def __init__(self, fn):
         super().__init__()
@@ -164,7 +171,7 @@ def image_noise(n, im_size):
     return torch.FloatTensor(n, im_size, im_size, 1).uniform_(0., 1.).cuda()
 
 def leaky_relu(p=0.2):
-    return nn.LeakyReLU(p, inplace=True)
+    return nn.LeakyReLU(p)
 
 def evaluate_in_chunks(max_batch_size, model, *args):
     split_args = list(zip(*list(map(lambda x: x.split(max_batch_size, dim=0), args))))
@@ -188,6 +195,14 @@ def slerp(val, low, high):
     res = (torch.sin((1.0 - val) * omega) / so).unsqueeze(1) * low + (torch.sin(val * omega) / so).unsqueeze(1) * high
     return res
 
+def warmup(start, end, max_steps, current_step):
+    if current_step > max_steps:
+        return end
+    return (end - start) * (current_step / max_steps) + start
+
+def log(t, eps = 1e-6):
+    return torch.log(t + eps)
+
 def cutmix_coordinates(height, width, alpha = 1.):
     lam = np.random.beta(alpha, alpha)
 
@@ -203,12 +218,14 @@ def cutmix_coordinates(height, width, alpha = 1.):
     return ((y0, y1), (x0, x1)), lam
 
 def cutmix(source, target, coors, alpha = 1.):
-    *_, h, w = source.shape
     source, target = map(torch.clone, (source, target))
     ((y0, y1), (x0, x1)), _ = coors
     source[:, :, y0:y1, x0:x1] = target[:, :, y0:y1, x0:x1]
     return source
 
+def mask_src_tgt(source, target, mask):
+    return source * mask + (1 - mask) * target
+
 # dataset
 
 def convert_rgb_to_transparent(image):
@@ -434,18 +451,23 @@ class DownBlock(nn.Module):
 class UpBlock(nn.Module):
     def __init__(self, input_channels, filters):
         super().__init__()
-
+        self.conv_res = nn.ConvTranspose2d(input_channels // 2, filters, 1, stride = 2)
         self.net = double_conv(input_channels, filters)
-        self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=False)
+        self.up = nn.Upsample(scale_factor = 2, mode='bilinear', align_corners=False)
+        self.input_channels = input_channels
+        self.filters = filters
 
     def forward(self, x, res):
+        *_, h, w = x.shape
+        conv_res = self.conv_res(x, output_size = (h * 2, w * 2))
         x = self.up(x)
         x = torch.cat((x, res), dim=1)
         x = self.net(x)
+        x = x + conv_res
         return x
 
 class Generator(nn.Module):
-    def __init__(self, image_size, latent_dim, network_capacity = 16, transparent = False, attn_layers = [], no_const = False, fmap_max = 512):
+    def __init__(self, image_size, latent_dim, network_capacity = 16, transparent = False, no_const = False, fmap_max = 512):
         super().__init__()
         self.image_size = image_size
         self.latent_dim = latent_dim
@@ -474,8 +496,7 @@ class Generator(nn.Module):
             not_last = ind != (self.num_layers - 1)
             num_layer = self.num_layers - ind
 
-            attn_fn = attn_and_ff(in_chan) if num_layer in attn_layers else None
-
+            attn_fn = attn_and_ff(in_chan)
             self.attns.append(attn_fn)
 
             block = GeneratorBlock(
@@ -509,9 +530,9 @@ class Generator(nn.Module):
         return rgb
 
 class Discriminator(nn.Module):
-    def __init__(self, image_size, network_capacity = 16, attn_layers = [], transparent = False, fmap_max = 512):
+    def __init__(self, image_size, network_capacity = 16, transparent = False, fmap_max = 512):
         super().__init__()
-        num_layers = int(log2(image_size) - 1)
+        num_layers = int(log2(image_size) - 2)
         num_init_filters = 3 if not transparent else 4
 
         blocks = []
@@ -519,6 +540,8 @@ class Discriminator(nn.Module):
 
         set_fmap_max = partial(min, fmap_max)
         filters = list(map(set_fmap_max, filters))
+        filters[-1] = filters[-2]
+
         chan_in_out = list(zip(filters[:-1], filters[1:]))
         chan_in_out = list(map(list, chan_in_out))
 
@@ -532,21 +555,24 @@ class Discriminator(nn.Module):
             block = DownBlock(in_chan, out_chan, downsample = is_not_last)
             down_blocks.append(block)
 
-            attn_fn = attn_and_ff(out_chan) if num_layer in attn_layers else None
-
+            attn_fn = attn_and_ff(out_chan)
             attn_blocks.append(attn_fn)
 
         self.down_blocks = nn.ModuleList(down_blocks)
         self.attn_blocks = nn.ModuleList(attn_blocks)
 
-        latent_dim = 2 * 2 * filters[-1]
+        last_chan = filters[-1]
 
-        self.to_logit = nn.Linear(latent_dim, 1)
+        self.to_logit = nn.Sequential(
+            leaky_relu(),
+            nn.AvgPool2d(image_size // (2 ** num_layers)),
+            Flatten(1),
+            nn.Linear(last_chan, 1)
+        )
 
-        self.conv = double_conv(filters[-1], filters[-1])
+        self.conv = double_conv(last_chan, last_chan)
 
         dec_chan_in_out = chan_in_out[:-1][::-1]
-
         self.up_blocks = nn.ModuleList(list(map(lambda c: UpBlock(c[1] * 2, c[0]), dec_chan_in_out)))
         self.conv_out = nn.Conv2d(3, 1, 1)
 
@@ -562,29 +588,28 @@ class Discriminator(nn.Module):
             if attn_block is not None:
                 x = attn_block(x)
 
-        enc_out = self.to_logit(x.flatten(1))
-
+        enc_out = self.to_logit(x)
         x = self.conv(x)
 
         for (up_block, res) in zip(self.up_blocks, residuals[:-1][::-1]):
             x = up_block(x, res)
 
         dec_out = self.conv_out(x)
-        return enc_out.squeeze(), dec_out
+        return enc_out.squeeze().sigmoid(), dec_out.sigmoid()
 
 class StyleGAN2(nn.Module):
-    def __init__(self, image_size, latent_dim = 512, fmap_max = 512, style_depth = 8, network_capacity = 16, transparent = False, fp16 = False, steps = 1, lr = 1e-4, ttur_mult = 2, attn_layers = [], no_const = False):
+    def __init__(self, image_size, latent_dim = 512, fmap_max = 512, style_depth = 8, network_capacity = 16, transparent = False, fp16 = False, steps = 1, lr = 1e-4, ttur_mult = 2, no_const = False):
         super().__init__()
         self.lr = lr
         self.steps = steps
         self.ema_updater = EMA(0.995)
 
         self.S = StyleVectorizer(latent_dim, style_depth)
-        self.G = Generator(image_size, latent_dim, network_capacity, transparent = transparent, attn_layers = attn_layers, no_const = no_const, fmap_max = fmap_max)
-        self.D = Discriminator(image_size, network_capacity, attn_layers = attn_layers, transparent = transparent, fmap_max = fmap_max)
+        self.G = Generator(image_size, latent_dim, network_capacity, transparent = transparent, no_const = no_const, fmap_max = fmap_max)
+        self.D = Discriminator(image_size, network_capacity, transparent = transparent, fmap_max = fmap_max)
 
         self.SE = StyleVectorizer(latent_dim, style_depth)
-        self.GE = Generator(image_size, latent_dim, network_capacity, transparent = transparent, attn_layers = attn_layers, no_const = no_const)
+        self.GE = Generator(image_size, latent_dim, network_capacity, transparent = transparent, no_const = no_const)
 
         # wrapper for augmenting all images going into the discriminator
         self.D_aug = AugWrapper(self.D, image_size)
@@ -633,7 +658,7 @@ class StyleGAN2(nn.Module):
         return x
 
 class Trainer():
-    def __init__(self, name, results_dir, models_dir, image_size, network_capacity, transparent = False, batch_size = 4, mixed_prob = 0.9, gradient_accumulate_every=1, lr = 2e-4, ttur_mult = 2, num_workers = None, save_every = 1000, trunc_psi = 0.6, fp16 = False, attn_layers = [], no_const = False, aug_prob = 0., dataset_aug_prob = 0., *args, **kwargs):
+    def __init__(self, name, results_dir, models_dir, image_size, network_capacity, transparent = False, batch_size = 4, mixed_prob = 0.9, gradient_accumulate_every=1, lr = 2e-4, ttur_mult = 2, num_workers = None, save_every = 1000, trunc_psi = 0.6, fp16 = False, no_const = False, aug_prob = 0., dataset_aug_prob = 0., cr_weight = 0.2, *args, **kwargs):
         self.GAN_params = [args, kwargs]
         self.GAN = None
 
@@ -646,8 +671,7 @@ class Trainer():
         self.image_size = image_size
         self.network_capacity = network_capacity
         self.transparent = transparent
-        
-        self.attn_layers = cast_list(attn_layers)
+
         self.no_const = no_const
         self.aug_prob = aug_prob
 
@@ -681,9 +705,11 @@ class Trainer():
         self.loader = None
         self.dataset_aug_prob = dataset_aug_prob
 
+        self.cr_weight = cr_weight
+
     def init_GAN(self):
         args, kwargs = self.GAN_params
-        self.GAN = StyleGAN2(lr = self.lr, ttur_mult = self.ttur_mult, image_size = self.image_size, network_capacity = self.network_capacity, transparent = self.transparent, attn_layers = self.attn_layers, fp16 = self.fp16, no_const = self.no_const, *args, **kwargs)
+        self.GAN = StyleGAN2(lr = self.lr, ttur_mult = self.ttur_mult, image_size = self.image_size, network_capacity = self.network_capacity, transparent = self.transparent, fp16 = self.fp16, no_const = self.no_const, *args, **kwargs)
 
     def write_config(self):
         self.config_path.write_text(json.dumps(self.config()))
@@ -693,13 +719,12 @@ class Trainer():
         self.image_size = config['image_size']
         self.network_capacity = config['network_capacity']
         self.transparent = config['transparent']
-        self.attn_layers = config.pop('attn_layers', [])
         self.no_const = config.pop('no_const', False)
         del self.GAN
         self.init_GAN()
 
     def config(self):
-        return {'image_size': self.image_size, 'network_capacity': self.network_capacity, 'transparent': self.transparent, 'attn_layers': self.attn_layers, 'no_const': self.no_const}
+        return {'image_size': self.image_size, 'network_capacity': self.network_capacity, 'transparent': self.transparent, 'no_const': self.no_const}
 
     def set_data_src(self, folder):
         self.dataset = Dataset(folder, self.image_size, transparent = self.transparent, aug_prob = self.dataset_aug_prob)
@@ -724,7 +749,10 @@ class Trainer():
         aug_prob   = self.aug_prob
 
         apply_gradient_penalty = self.steps % 4 == 0
-        apply_path_penalty = self.steps % 32 == 0        
+        apply_path_penalty = self.steps % 32 == 0
+
+        cutmix_prob = warmup(0, 0.5, 30000, self.steps)
+        apply_cutmix = random() < cutmix_prob
 
         backwards = partial(loss_backwards, self.fp16)
 
@@ -749,22 +777,38 @@ class Trainer():
             (real_enc_out, real_dec_out), real_aug_images = self.GAN.D_aug(real_images, prob = aug_prob)
 
             cutmix_coors = cutmix_coordinates(image_size, image_size)
-            cutmixed_images = cutmix(real_aug_images, fake_aug_images, cutmix_coors)
-            cutmix_enc_out, cutmix_dec_out = self.GAN.D(cutmixed_images)
 
-            cr_cutmix_dec_out = cutmix(real_dec_out, fake_dec_out, cutmix_coors)
-            cr_loss = F.mse_loss(cutmix_dec_out, cr_cutmix_dec_out)
-            self.last_cr_loss = cr_loss.clone().detach().item()
+            mask = cutmix(
+                torch.ones_like(real_dec_out),
+                torch.zeros_like(real_dec_out),
+                cutmix_coors
+            )
 
-            enc_divergence = (F.relu(1 + real_enc_out) + F.relu(1 - fake_enc_out)).mean()
-            dec_divergence = (F.relu(1 + real_dec_out) + F.relu(1 - fake_dec_out)).mean()
+            if random() > 0.5:
+                mask = 1 - mask
+
+            cutmix_images = mask_src_tgt(real_aug_images, fake_aug_images, mask)
+            cutmix_enc_out, cutmix_dec_out = self.GAN.D(cutmix_images)
+
+            enc_divergence = -(log(1 - real_enc_out) + log(fake_enc_out)).mean()
+            dec_divergence = -(log(1 - real_dec_out) + log(fake_dec_out)).mean()
             divergence = enc_divergence + dec_divergence
 
             disc_loss = divergence
-            disc_loss = disc_loss + cr_loss
+
+            if apply_cutmix:
+                cutmix_enc_divergence = -log(cutmix_enc_out).mean()
+                cutmix_dec_divergence =  F.binary_cross_entropy(cutmix_dec_out.flatten(1), (1 - mask).flatten(1), reduction='mean')
+                disc_loss = disc_loss + cutmix_enc_divergence + cutmix_dec_divergence
+
+                cr_cutmix_dec_out = mask_src_tgt(real_dec_out, fake_dec_out, mask)
+                cr_loss = F.mse_loss(cutmix_dec_out, cr_cutmix_dec_out) * self.cr_weight
+                self.last_cr_loss = cr_loss.clone().detach().item()
+
+                disc_loss = disc_loss + cr_loss
 
             if apply_gradient_penalty:
-                gp = gradient_penalty(real_images, real_dec_out)
+                gp = gradient_penalty(real_images, real_enc_out)
                 self.last_gp_loss = gp.clone().detach().item()
                 disc_loss = disc_loss + gp
 
@@ -790,7 +834,7 @@ class Trainer():
 
             generated_images = self.GAN.G(w_styles, noise)
             (fake_enc_output, fake_dec_output), _ = self.GAN.D_aug(generated_images, prob = aug_prob)
-            loss = fake_enc_output.mean() + fake_dec_output.mean()
+            loss = -log(1 - fake_enc_output).mean() - log(1 - fake_dec_output).mean()
             gen_loss = loss
 
             if apply_path_penalty:
